@@ -3,7 +3,16 @@ __license__ = "GPLv2"
 __maintainer__ = "Joseph Ryan"
 __email__ = "jr@aphyt.com"
 
+import concurrent.futures
+import pickle
+import socket
+import struct
+import sys
+import threading
+from signal import signal, SIGINT
+
 from aphyt.eip import *
+import logging
 
 
 class VariableTypeObjectReply(CIPReply):
@@ -155,6 +164,7 @@ class SimpleDataSegmentRequest:
     A simple data segment request is the Omron specific format for requesting data that will not fit
     in a single CIP message. This is used for strings, arrays and structures
     """
+
     def __init__(self, offset, size):
         self.simple_data_type_code = b'\x80'
         self.segment_length = b'\x03'  # Fixed and in words
@@ -169,9 +179,11 @@ class SimpleDataSegmentRequest:
 
 class NSeries:
     """
-    Concrete implementation of EIP abstract base class that implements Omron specific Ethernet/IP
-    services in addition to the common Ethernet/IP services of the parent class and the CIP specific
-    items in the grandparent class
+    NSeries
+    ~~~~~~~
+
+    Class that implements Omron N-Series specific Ethernet/IP services in addition to Ethernet/IP services
+    and CIP services common to most Ethernet/IP devices
     """
     MAXIMUM_LENGTH = 502  # UCMM maximum length is 502 bytes
 
@@ -202,7 +214,7 @@ class NSeries:
     def update_derived_data_type_dictionary(self):
         # ToDo get the derived data types in such  a way they are easy to use
         number_of_entries = self._get_number_of_derived_data_types()
-        for index in range(1, number_of_entries+1):
+        for index in range(1, number_of_entries + 1):
             request_path = eip.address_request_path_segment(class_id=b'\x6c', instance_id=index.to_bytes(2, 'little'))
             reply = VariableTypeObjectReply(
                 self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
@@ -258,6 +270,32 @@ class NSeries:
                     self.connected_cip_dispatcher.user_variables.update({variable: variable_cip_datatype})
             instance_id = instance_id + 1
 
+    def save_current_dictionary(self, file_name: str):
+        with (open(file_name, "wb")) as f:
+            pickle.dump(self.connected_cip_dispatcher.variables, f)
+
+    def load_dictionary_file(self, file_name: str):
+        with (open(file_name, "rb")) as f:
+            self.connected_cip_dispatcher.variables = pickle.load(f)
+
+    def load_dictionary_file_if_present(self, file_name: str):
+        """
+        If the supplied file name exists, load it as the controller's dictionary, and if not, pull it
+        from the controller and save the dictionary as the supplied file name.
+
+        This allows the programmer to use a file instead of network operation to quickly populate the variable
+        dictionary, and if the controller's variable dictionary changes, the user can get the new values
+        by simply deleting the file and it will be repopulated on the next run.
+        :param file_name:
+        :return:
+        """
+        try:
+            with (open(file_name, "rb")) as f:
+                self.connected_cip_dispatcher.variables = pickle.load(f)
+        except IOError as err:
+            self.update_variable_dictionary()
+            self.save_current_dictionary(file_name)
+
     def _get_data_instance(self, cip_datatype_instance: CIPDataType) -> CIPDataType:
         """
         This method is to get an instance of a CIP data type from its type definition.
@@ -308,7 +346,8 @@ class NSeries:
                     int.from_bytes(member_cip_datatype_instance.next_instance_id, 'little')
             return cip_datatype_instance
         else:
-            cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(cip_datatype_instance.data_type_code())()
+            cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(
+                cip_datatype_instance.data_type_code())()
             return cip_datatype_instance
 
     def _get_member_instance(self, member_instance_id: int) -> CIPDataType:
@@ -319,7 +358,8 @@ class NSeries:
         :return:
         """
         variable_type_object_reply = self._get_variable_type_object(member_instance_id)
-        cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(variable_type_object_reply.cip_data_type)()
+        cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(
+            variable_type_object_reply.cip_data_type)()
         cip_datatype_instance.variable_type_name = str(variable_type_object_reply.variable_type_name, 'utf-8')
         cip_datatype_instance.size = variable_type_object_reply.size_in_memory
         cip_datatype_instance.next_instance_id = variable_type_object_reply.next_instance_id
@@ -393,6 +433,17 @@ class NSeries:
             request_data = CIPCommonFormat(cip_datatype_object.data_type_code(), data=cip_datatype_object.data)
             self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
 
+    def verified_write_variable(self, variable_name: str, data, retry_count: int = 2):
+        remaining_retry = retry_count
+        temp_data = None
+        while remaining_retry >= 0 and temp_data != data:
+            self.write_variable(variable_name, data)
+            temp_data = self.read_variable(variable_name)
+            remaining_retry += -1
+        if remaining_retry < 0 and temp_data != data:
+            logging.warning('Write Operation could not be completed within specified retry count')
+            raise IOError('Write Operation could not be completed within specified retry count')
+
     def _multi_message_variable_read(self, cip_datatype_object: CIPDataType, offset=0):
         """
         This method is to read data that does not fit into a single CIP message
@@ -509,7 +560,8 @@ class NSeries:
         """
         request_path = address_request_path_segment(
             class_id=b'\x6c', instance_id=instance_id.to_bytes(2, 'little'))
-        variable_type_object_reply = VariableTypeObjectReply(self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
+        variable_type_object_reply = VariableTypeObjectReply(
+            self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
         return variable_type_object_reply
 
     def _get_number_of_derived_data_types(self) -> int:
@@ -527,7 +579,8 @@ class NSeries:
         """
         request_path = address_request_path_segment(
             class_id=b'\x6b', instance_id=instance_id.to_bytes(2, 'little'))
-        variable_object_reply = VariableObjectReply(self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
+        variable_object_reply = VariableObjectReply(
+            self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
         return variable_object_reply
 
     def _get_variable_list(self):
@@ -576,3 +629,233 @@ class NSeries:
 
     def _cip_array_write(self):
         pass
+
+
+class EIPConnectionStatus:
+    def __init__(self):
+        self.connected = False
+        self.connecting = False
+        self.reconnecting = False
+        self._has_session = False
+        self.persist_session = False
+        self._session_status_observers = []
+        self.keep_alive = False
+        self.keep_alive_running = False
+
+    def bind_to_session_status(self, callback):
+        """
+        Observer Pattern: Allow widgets to bind a callback function that will act on reestablished session
+        :param callback:
+        :return:
+        """
+        self._session_status_observers.append(callback)
+
+    @property
+    def has_session(self):
+        return self._has_session
+
+    @has_session.setter
+    def has_session(self, value):
+        self._has_session = value
+        for callback in self._session_status_observers:
+            callback()
+
+
+class MonitoredVariable:
+    def __init__(self, dispatcher: "NSeriesThreadDispatcher", variable_name,
+                 refresh_time: float = 0.05):
+        self.variable_name = variable_name
+        self.dispatcher = dispatcher
+        self.monitored_variable_observers = []
+        self.refresh_timer = threading.Timer(refresh_time, self.update)
+        self.refresh_time = refresh_time
+        self._value = None
+        self.update()
+
+    def bind_to_value(self, callback):
+        """
+        Observer Pattern: Allow widgets to bind a callback function that will act on value change
+        :param callback:
+        :return:
+        """
+        self.monitored_variable_observers.append(callback)
+
+    @property
+    def value(self):
+        return self._value
+
+    @value.setter
+    def value(self, value):
+        self.dispatcher.verified_write_variable(self.variable_name, value)
+        self._value = value
+        for callback in self.monitored_variable_observers:
+            callback()
+
+    def update(self):
+        temp_value = self._value
+        self._value = self.dispatcher.read_variable(self.variable_name)
+        if temp_value != self._value:
+            for callback in self.monitored_variable_observers:
+                callback()
+        self.refresh_timer = threading.Timer(self.refresh_time, self.update)
+        self.refresh_timer.daemon = True
+        self.refresh_timer.start()
+
+    def start(self):
+        self.refresh_timer.start()
+
+    def cancel(self):
+        self.refresh_timer.cancel()
+
+
+class NSeriesThreadDispatcher:
+    def __init__(self):
+        self._instance = NSeries()
+        self._host = None
+        self.message_timeout = 0.5
+        self.executor = None
+        self.futures = []
+        self.status_integer = 0
+        self.services = None
+        self.monitored_variable_dictionary = {}
+        self.connection_status = EIPConnectionStatus()
+
+    def add_monitored_variable(self, monitored_variable: MonitoredVariable):
+        self.monitored_variable_dictionary[monitored_variable.variable_name] = monitored_variable
+
+    def _execute_eip_command(self, command, *args, **kwargs):
+        if self.connection_status.connected:
+            future = self.executor.submit(command, *args, **kwargs)
+            try:
+                result = future.result()
+                return result
+            except socket.error as err:
+                print(f'{err} occurred in _execute_eip_command')
+                self.connection_status.connected = False
+                self.connection_status.keep_alive_running = False
+                self.executor.shutdown()
+                self._reconnect()
+        else:
+            self._reconnect()
+
+    def retry_command(self, command, retry_time: float = 0.05, *args, **kwargs):
+        reply = command(*args, **kwargs)
+        if not reply:
+            delay = threading.Timer(retry_time, self.retry_command, command,
+                                    *args, **dict(**kwargs, retry_time=retry_time))
+            delay.daemon = True
+            delay.start()
+        return reply
+
+    def _reconnect(self, already_executing=False):
+        if not already_executing and not self.connection_status.reconnecting:
+            self.connection_status.reconnecting = True
+            self._reconnect(already_executing=True)
+        elif self.connection_status.reconnecting and already_executing:
+            if not self.connection_status.connected:
+                try:
+                    self.close_explicit()
+                except Exception as err:
+                    print(f'{err} occurred in close_explicit')
+                self.connect_explicit(self._host)
+                self._reconnect(already_executing=True)
+            if (self.connection_status.persist_session
+                    and self.connection_status.connected
+                    and not self.connection_status.has_session):
+                self.register_session()
+                self._reconnect(already_executing=True)
+            elif not self.connection_status.persist_session and not self.connection_status.keep_alive:
+                self.connection_status.reconnecting = False
+            if self.connection_status.keep_alive and self.connection_status.has_session:
+                self.start_keep_alive()
+            elif not self.connection_status.keep_alive:
+                self.connection_status.reconnecting = False
+
+    def start_keep_alive(self, keep_alive_time: float = 0.05):
+        self.connection_status.keep_alive = True
+        if (self.connection_status.connected
+                and self.connection_status.keep_alive
+                and self.connection_status.has_session):
+            self.connection_status._keep_alive_running = True
+            self.services = self._execute_eip_command(self._instance.connected_cip_dispatcher.list_services, '')
+            delay = threading.Timer(keep_alive_time, self.start_keep_alive, [keep_alive_time])
+            delay.daemon = True
+            delay.start()
+
+    def connect_explicit(self, host: str, retry_time: float = 1.0):
+        self.connection_status.connecting = True
+        if not self.connection_status.connected:
+            self._host = host
+            try:
+                self._instance.connect_explicit(host=self._host)
+                self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
+                self.connection_status.connected = self._instance.connected_cip_dispatcher.is_connected_explicit
+                self.connection_status.connecting = False
+            except socket.error as err:
+                print(f"Failed to connect, trying again in {retry_time} seconds")
+                delay = threading.Timer(retry_time, self.connect_explicit, [host, retry_time])
+                delay.start()
+
+    def register_session(self, retry_time: float = 1.0):
+        self._instance.register_session()
+        session_id = self._instance.connected_cip_dispatcher.session_handle_id
+        if session_id != b'\x00\x00\x00\x00\x00\x00\x00\x00':
+            self.connection_status.has_session = True
+        else:
+            print(f"Failed to register session, trying again in {retry_time} seconds")
+            delay = threading.Timer(retry_time, self.register_session, [retry_time])
+            delay.start()
+        self.connection_status.persist_session = True
+
+    def unregister_session(self):
+        self.connection_status.has_session = False
+        self.connection_status.persist_session = False
+
+    def update_variable_dictionary(self):
+        self._instance.update_variable_dictionary()
+
+    def save_current_dictionary(self, file_name: str):
+        self._instance.save_current_dictionary(file_name)
+
+    def load_dictionary_file(self, file_name: str):
+        self._instance.load_dictionary_file(file_name)
+
+    def load_dictionary_file_if_present(self, file_name: str):
+        self._instance.load_dictionary_file_if_present(file_name)
+
+    def read_variable(self, variable_name: str):
+        try:
+            return self._execute_eip_command(self._instance.read_variable, variable_name)
+        except struct.error as err:
+            pass
+
+    def write_variable(self, variable_name: str, data):
+        return self._execute_eip_command(self._instance.write_variable, variable_name, data)
+
+    def verified_write_variable(self, variable_name: str, data):
+        try:
+            return self._execute_eip_command(self._instance.verified_write_variable, variable_name, data)
+        except struct.error as error:
+            raise error
+
+    def close_explicit(self):
+        self.connection_status.connected = False
+        self.connection_status.has_session = False
+        self.connection_status._keep_alive_running = False
+        self.executor.shutdown(wait=True)
+        for monitored in self.monitored_variable_dictionary:
+            self.monitored_variable_dictionary[monitored].cancel()
+        if self._instance.connected_cip_dispatcher.is_connected_explicit:
+            self._instance.close_explicit()
+
+    def __enter__(self):
+        signal(SIGINT, self._sigint_handler)
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close_explicit()
+
+    def _sigint_handler(self, signal_received, frame):
+        print('Program interrupted with Ctrl+C')
+        self.__exit__(None, None, None)
+        sys.exit(0)
